﻿using System.Diagnostics;
using System.Windows.Forms;
using System.Drawing;
using System.Threading;
using System;
using System.Reflection;
using Microsoft.Win32;
using System.IO;

namespace GraphicsSwitcher
{
    public class Switcher : IDisposable
    {
        /// <summary>
        /// Hidden form that receives power messages.
        /// </summary>
        private readonly PowerMsgForm PwrMsgForm;

        /// <summary>
        /// Count of switching retries, if it does not succeed.
        /// </summary>
        public const int RetryCount = 30;

        /// <summary>
        /// Time in ms between two switching attempts.
        /// </summary>
        public const int RetryLatency = 3000;

        /// <summary>
        /// Name of the source for event log entries
        /// </summary>
        public const string EventLogSourceName = "GraphicsSwitcher";

        public event GraphicsSwitchedDelegate OnGraphicsSwitch;


        /// <summary>
        /// Constructor.
        /// </summary>
        public Switcher()
        {
            //Create hidden form that receives power messages
            PwrMsgForm = new PowerMsgForm();
            PwrMsgForm.OnGraphicsSwitch += new GraphicsSwitchedDelegate(PwrMsgForm_OnGraphicsSwitch);

            PwrMsgForm.Size = new Size(0, 0);
            PwrMsgForm.FormBorderStyle = FormBorderStyle.None;
            PwrMsgForm.Show();
            PwrMsgForm.Hide();

            //Create event log source
            try
            {
                if (!EventLog.SourceExists(EventLogSourceName))
                    EventLog.CreateEventSource(EventLogSourceName, EventLogSourceName);
            }
            catch (Exception) {} //Will happen when the registry is not accessible with the given rights
                                 //No problem, as event log is only used for debugging (with administrative rights)
        }

        void PwrMsgForm_OnGraphicsSwitch(GraphicsAdapterType activeAdapter)
        {
            if (OnGraphicsSwitch != null)
                OnGraphicsSwitch(activeAdapter);
        }

        public void Dispose()
        {
            PwrMsgForm.Dispose();
        }
    }

    public enum GraphicsAdapterType
    {
        Integrated,
        Dedicated,
    }
    public delegate void GraphicsSwitchedDelegate(GraphicsAdapterType activeAdapter);


    public class PowerMsgForm : Form
    {
        public event GraphicsSwitchedDelegate OnGraphicsSwitch;

        /// <summary>
        /// Overridden window procedure.
        /// </summary>
        /// <param name="m">Window message.</param>
        protected override void WndProc(ref Message m)
        {
            if(m.Msg == 1) //WM_CREATE
            {
                //Swith on program startup
                SetPowerState(SystemInformation.PowerStatus.PowerLineStatus == PowerLineStatus.Online);
            }

            if (m.Msg == 536) //WM_POWERBROADCAST
            {
                if(m.WParam.ToInt32() == 10) //PBT_APMPOWERSTATUSCHANGE
                {
                    //Set graphics adapter depending on power status
                    SetPowerState(SystemInformation.PowerStatus.PowerLineStatus == PowerLineStatus.Online);
                }
                return;
            }
            base.WndProc(ref m);
        }

        /// <summary>
        /// Time of the most recently started SetPowerState thread.
        /// </summary>
        private static DateTime MostRecentSetTime = DateTime.MinValue;

        /// <summary>
        /// Object used to lock MostRecentSetTime (as it is no reference object itself.
        /// </summary>
        private static readonly object MostRecentSetTimeLockObj = new object();

        /// <summary>
        /// Activates a graphics adapter.
        /// </summary>
        /// <param name="setPerformanceState">True, if the performance adapter is to be activated, false not.</param>
        private void SetPowerState(bool setPerformanceState)
        {
            Thread setThread = new Thread(SetPowerStateThread);
            setThread.Start(setPerformanceState);
        }

        /// <summary>
        /// Thread: Activates a graphics adapter.
        /// </summary>
        /// <param name="setPerformanceState">True, if the performance adapter is to be activated, false not.</param>
        private void SetPowerStateThread(object setPerformanceState)
        {
            if(!(setPerformanceState is bool))
                throw new ArgumentException("setPerformanceState must be of type bool.");

            //Save thread start time and set as most recent set thread
            DateTime setTime = DateTime.Now;
            lock (MostRecentSetTimeLockObj)
                MostRecentSetTime = setTime;


            //Get PWMUICtl.dll path
            string pwmAssemblyPath;
            try
            {
                string serviceImagePath = (string)Registry.LocalMachine.OpenSubKey(@"SYSTEM\CurrentControlSet\services\Power Manager DBC Service").GetValue("ImagePath");
                serviceImagePath = serviceImagePath.Trim('\"');
                pwmAssemblyPath = serviceImagePath.Substring(0, serviceImagePath.LastIndexOf('\\') + 1) + "PWMUICtl.dll";
            }
            catch (Exception)
            {
                WriteErrorToEventLog("Could not get path of PWMUICtl.dll.");
                return;
            }


            //Load PWMUICtl assembly
            Assembly pwmAssembly;
            try
            {
                pwmAssembly = Assembly.LoadFrom(pwmAssemblyPath);
            }
            catch (Exception)
            {
                WriteErrorToEventLog(string.Format("Could not load \"{0}\".", pwmAssemblyPath));
                return;
            }

            //Find HybridGraphics type
            Type hybridGraphicsType = null;
            foreach(Module pwmModule in pwmAssembly.GetModules())
            {
                try
                {
                    hybridGraphicsType = pwmModule.GetType("Lenovo.PowerManager.PWMUICtl.HybridGraphics");
                }
                catch (Exception) {} //If an exception occurs the type will not be set and the function will be aborted some lines later
                if(hybridGraphicsType != null)
                    break;
            }
            if (hybridGraphicsType == null)
            {
                WriteErrorToEventLog("Could find HybridGraphics type in PWMUICtl.dll.");
                return;
            }

            //Instanciate HybridGraphics
            object hybridGraphics;
            try
            {
                hybridGraphics = Activator.CreateInstance(hybridGraphicsType);
            }
            catch (Exception)
            {
                WriteErrorToEventLog("Could not instanciate HybridGraphics.");
                return;
            }


            //If the system is not capable of switching - return
            if (!(bool)hybridGraphicsType.GetMethod("IsCapable").Invoke(hybridGraphics, null))
            {
                WriteErrorToEventLog("Graphics adapter is not capable of dynamic switching.");
                return;
            }

            //Disable (most) warning messages displayed to user
            hybridGraphicsType.GetProperty("ShowWarningMessage").SetValue(hybridGraphics, true, null);


            int retries = 0;
            MethodInfo isPerformanceGfxEnableMethodInfo = hybridGraphicsType.GetMethod("IsPerformanceGfxEnable");
            MethodInfo activateAdapterMethodInfo = hybridGraphicsType.GetMethod((bool) setPerformanceState ? "EnablePerformanceGfx" : "DisablePerformanceGfx");
            //Retry until switching succeeds (or retries exceed)
            while ((bool)isPerformanceGfxEnableMethodInfo.Invoke(hybridGraphics, null) != (bool)setPerformanceState)
            {
                //Wait before first as well as between switches
                Thread.Sleep(Switcher.RetryLatency);

                lock (MostRecentSetTimeLockObj)
                {
                    //Check if other set power thread has been started - if so, quit this one
                    if (MostRecentSetTime > setTime)
                        return;

                    //Active correct graphics adapter
                    activateAdapterMethodInfo.Invoke(hybridGraphics, null);
                }

                //Increment retry counter, break if needed
                if (retries >= Switcher.RetryCount)
                    break;
                retries++;
            }

            //Fire graphics switch event if necessary
            if ((bool)isPerformanceGfxEnableMethodInfo.Invoke(hybridGraphics, null) == (bool)setPerformanceState)
            {
                if (OnGraphicsSwitch != null)
                    OnGraphicsSwitch((bool) setPerformanceState ? GraphicsAdapterType.Dedicated : GraphicsAdapterType.Integrated);
            }
        }

        private static void WriteErrorToEventLog(string message)
        {
            try
            {
                using (EventLog eventLog = new EventLog(Switcher.EventLogSourceName))
                {
                    eventLog.EnableRaisingEvents = false;
                    eventLog.WriteEntry(message, EventLogEntryType.Error);
                }
            }
            catch (Exception) { } //Application not startet with sufficiant rights. Silent error.
        }
    }
}
